@@ -1,6 +1,6 @@
 ;;; Continuation-passing style (CPS) intermediate language (IL)
 
-;; Copyright (C) 2013, 2014, 2015 Free Software Foundation, Inc.
+;; Copyright (C) 2013-2021, 2023 Free Software Foundation, Inc.
 
 ;;;; This library is free software; you can redistribute it and/or
 ;;;; modify it under the terms of the GNU Lesser General Public
@@ -80,10 +80,19 @@ sites."
                                      (causes-effect? fx &allocation))
                                 (values (intset-add! known k) unknown)
                                 (values known (intset-add! unknown k)))))
+                         (($ $kargs _ _ (or ($ $branch) ($ $switch)
+                                            ($ $prompt) ($ $throw)))
+                          ;; Branches, switches, and prompts pass no
+                          ;; values to their continuations, and throw
+                          ;; terms don't continue at all.
+                          (values known unknown))
                          (($ $kreceive arity kargs)
                           (values known (intset-add! unknown kargs)))
-                         (($ $kfun src meta self tail clause)
-                          (values known unknown))
+                         (($ $kfun src meta self tail entry)
+                          (values known
+                                  (if entry
+                                      (intset-add! unknown entry)
+                                      unknown)))
                          (($ $kclause arity body alt)
                           (values known (intset-add! unknown body)))
                          (($ $ktail)
@@ -129,7 +138,9 @@ sites."
          (values live-labels live-vars))
         (($ $fun body)
          (values (intset-add live-labels body) live-vars))
-        (($ $closure body)
+        (($ $const-fun body)
+         (values (intset-add live-labels body) live-vars))
+        (($ $code body)
          (values (intset-add live-labels body) live-vars))
         (($ $rec names vars (($ $fun kfuns) ...))
          (let lp ((vars vars) (kfuns kfuns)
@@ -142,19 +153,17 @@ sites."
                       (intset-add live-labels kfun)
                       live-labels)
                   live-vars)))))
-        (($ $prompt escape? tag handler)
-         (values live-labels (adjoin-var tag live-vars)))
         (($ $call proc args)
          (values live-labels (adjoin-vars args (adjoin-var proc live-vars))))
         (($ $callk kfun proc args)
          (values (intset-add live-labels kfun)
-                 (adjoin-vars args (adjoin-var proc live-vars))))
-        (($ $primcall name args)
+                 (adjoin-vars args (if proc
+                                       (adjoin-var proc live-vars)
+                                       live-vars))))
+        (($ $calli args callee)
+         (values live-labels (adjoin-var callee (adjoin-vars args live-vars))))
+        (($ $primcall name param args)
          (values live-labels (adjoin-vars args live-vars)))
-        (($ $branch k ($ $primcall name args))
-         (values live-labels (adjoin-vars args live-vars)))
-        (($ $branch k ($ $values (arg)))
-         (values live-labels (adjoin-var arg live-vars)))
         (($ $values args)
          (values live-labels
                  (match (cont-defs k)
@@ -175,8 +184,6 @@ sites."
           (or
            ;; No defs; perhaps continuation is $ktail.
            (not defs)
-           ;; We don't remove branches.
-           (match exp (($ $branch) #t) (_ #f))
            ;; Do we have a live def?
            (any-var-live? defs live-vars)
            ;; Does this expression cause all effects?  If so, it's
@@ -191,10 +198,14 @@ sites."
            (and (causes-effect? fx &write)
                 (match exp
                   (($ $primcall
-                      (or 'vector-set! 'vector-set!/immediate
+                      (or 'scm-set! 'scm-set!/tag 'scm-set!/immediate
+                          'word-set! 'word-set!/immediate
+                          'vector-set! 'vector-set!/immediate
                           'set-car! 'set-cdr!
-                          'box-set!)
-                      (obj . _))
+                          'box-set!
+                          'struct-set!
+                          'closure-set!)
+                      _ (obj . _))
                    (or (var-live? obj live-vars)
                        (not (intset-ref known-allocations obj))))
                   (_ #t)))))
@@ -204,6 +215,44 @@ sites."
         ;; Still dead.
         (values live-labels live-vars))))
 
+    ;; Note, this is for $branch or $switch.
+    (define (visit-branch label kf kt* args live-labels live-vars)
+      (define (next-live-term k)
+        ;; FIXME: For a chain of dead branches, this is quadratic.
+        (let lp ((seen empty-intset) (k k))
+          (cond
+           ((intset-ref live-labels k) k)
+           ((intset-ref seen k) k)
+           (else
+            (match (intmap-ref conts k)
+              (($ $kargs _ _ ($ $continue k*))
+               (lp (intset-add seen k) k*))
+              (_ k))))))
+      (define (distinct-continuations?)
+        (let ((kf' (next-live-term kf)))
+          (let lp ((kt* kt*))
+            (match kt*
+              (() #f)
+              ((kt . kt*)
+               (cond
+                ((or (eqv? kf kt)
+                     (eqv? kf' (next-live-term kt)))
+                 (lp kt*))
+                (else #t)))))))
+      (cond
+       ((intset-ref live-labels label)
+        ;; Branch live already.
+        (values live-labels (adjoin-vars args live-vars)))
+       ((or (causes-effect? (intmap-ref effects label) &type-check)
+            (distinct-continuations?))
+        ;; The branch is live if its continuations are not the same, or
+        ;; if the branch itself causes type checks.
+        (values (intset-add live-labels label)
+                (adjoin-vars args live-vars)))
+       (else
+        ;; Still dead.
+        (values live-labels live-vars))))
+
     (define (visit-fun label live-labels live-vars)
       ;; Visit uses before definitions.
       (postorder-fold-local-conts2
@@ -211,12 +260,28 @@ sites."
          (match cont
            (($ $kargs _ _ ($ $continue k src exp))
             (visit-exp label k exp live-labels live-vars))
+           (($ $kargs _ _ ($ $branch kf kt src op param args))
+            (visit-branch label kf (list kt) args live-labels live-vars))
+           (($ $kargs _ _ ($ $switch kf kt* src arg))
+            (visit-branch label kf kt* (list arg) live-labels live-vars))
+           (($ $kargs _ _ ($ $prompt k kh src escape? tag))
+            ;; Prompts need special elision passes that would contify
+            ;; aborts and remove corresponding "unwind" primcalls.
+            (values (intset-add live-labels label)
+                    (adjoin-var tag live-vars)))
+           (($ $kargs _ _ ($ $throw src op param args))
+            ;; A reachable "throw" is always live.
+            (values (intset-add live-labels label)
+                    (adjoin-vars args live-vars)))
            (($ $kreceive arity kargs)
             (values live-labels live-vars))
            (($ $kclause arity kargs kalt)
             (values live-labels (adjoin-vars (cont-defs kargs) live-vars)))
-           (($ $kfun src meta self)
-            (values live-labels (adjoin-var self live-vars)))
+           (($ $kfun src meta self tail entry)
+            (values live-labels
+                    (adjoin-vars
+                     (or (and entry (cont-defs entry)) '())
+                     (if self (adjoin-var self live-vars) live-vars))))
            (($ $ktail)
             (values live-labels live-vars))))
        conts label live-labels live-vars))
@@ -272,7 +337,7 @@ sites."
              (($ $fun body)
               (values cps
                       term))
-             (($ $closure body nfree)
+             (($ $const-fun body)
               (values cps
                       term))
              (($ $rec names vars funs)
@@ -312,7 +377,22 @@ sites."
                  (values cps term)))))
            (values cps
                    (build-term
-                     ($continue k src ($values ()))))))))
+                     ($continue k src ($values ()))))))
+      (($ $branch kf kt src op param args)
+       (if (label-live? label)
+           (values cps term)
+           ;; Dead branches continue to the same continuation
+           ;; (eventually).
+           (values cps (build-term ($continue kf src ($values ()))))))
+      (($ $switch kf kt* src arg)
+       ;; Same as in $branch case.
+       (if (label-live? label)
+           (values cps term)
+           (values cps (build-term ($continue kf src ($values ()))))))
+      (($ $prompt)
+       (values cps term))
+      (($ $throw)
+       (values cps term))))
   (define (visit-cont label cont cps)
     (match cont
       (($ $kargs names vars term)
